from typing import Optional

from local_tuya.domoticz.units.base import Unit, UnitValues
from local_tuya.domoticz.units.value_preprocessors import ValuePreprocessor


def temperature_unit(
    id_: int,
    name: str,
    image: int,
    value_preprocessor: Optional[ValuePreprocessor[float]] = None,
) -> Unit[float]:
    def _to_unit_values(value: float) -> UnitValues:
        value = value if not value_preprocessor else value_preprocessor(value)
        return UnitValues(n_value=1, s_value=str(round(value, 1)))

    return Unit(
        id_=id_,
        type_="Temperature",
        name=name,
        image=image,
        to_unit_values=_to_unit_values,
    )



import binascii
import json
import logging
import struct
from typing import Callable, ClassVar, Dict, Optional, Tuple, Type

from local_tuya.errors import DecodeResponseError, LocalTuyaError, ResponseError
from local_tuya.protocol.config import ProtocolConfig, Version
from local_tuya.protocol.message.handlers.crypto import AESCipher
from local_tuya.protocol.message.handlers.handler import MessageHandler
from local_tuya.protocol.message.messages import (
    Command,
    HeartbeatCommand,
    HeartbeatResponse,
    Payload,
    Response,
    StateCommand,
    StateResponse,
    StatusResponse,
    UpdateCommand,
    UpdateResponse,
)

logger = logging.getLogger(__name__)


class V33MessageHandler(MessageHandler):
    PREFIX: ClassVar[int] = 0x000055AA
    SUFFIX: ClassVar[int] = 0x0000AA55
    VERSION_HEADER: ClassVar[bytes] = 12 * b"\x00"

    # Prefix + sequence num + command + payload length.
    HEADER_FMT: ClassVar[str] = ">4I"
    RETURN_CODE_FMT: ClassVar[str] = ">I"
    # Hash + suffix.
    END_FMT: ClassVar[str] = ">2I"

    # Messages.
    COMMANDS: ClassVar[Dict[Type[Command], int]] = {
        UpdateCommand: 7,
        HeartbeatCommand: 9,
        StateCommand: 10,
    }
    RESPONSES: ClassVar[
        Dict[int, Callable[[Optional[Payload], Optional[ResponseError]], Response]]
    ] = {
        7: lambda p, e: UpdateResponse(e),
        8: StatusResponse,
        9: lambda p, e: HeartbeatResponse(e),
        10: StateResponse,
    }
    # Corresponding to responses.
    COMMAND_CLASSES: ClassVar[Dict[int, Type[Command]]] = {
        7: UpdateCommand,
        9: HeartbeatCommand,
        10: StateCommand,
    }

    def __init__(self, config: ProtocolConfig):
        self._cfg = config
        self._cipher = AESCipher(config.key)
        self._version_header = config.version + self.VERSION_HEADER

    @staticmethod
    def from_config(config: ProtocolConfig) -> Optional["MessageHandler"]:
        if config.version is Version.v33:
            return V33MessageHandler(config)

    def pack(self, sequence_number: int, command: Command) -> bytes:
        if type(command) not in self.COMMANDS:
            raise LocalTuyaError(f"unknown command {command}")
        payload = command.payload or {}
        # TODO: check if needed.
        # payload.update({
        #     # "gwId": self._cfg.id,
        #     # "devId": self._cfg.id,
        # })
        # if not isinstance(command, HeartBeatCommand):
        #     payload.update({
        #         # "uid": self._cfg.id,
        #         "t": int(time.time()),
        #     })
        encrypted = self._cipher.encrypt(
            json.dumps(payload, separators=(",", ":")).encode()
        )
        if isinstance(command, UpdateCommand):
            full_payload = self._version_header + encrypted
        else:
            full_payload = encrypted

        data = (
            struct.pack(
                self.HEADER_FMT,
                self.PREFIX,
                sequence_number,
                self.COMMANDS[type(command)],
                len(full_payload) + struct.calcsize(self.END_FMT),
            )
            + full_payload
        )
        data += struct.pack(
            self.END_FMT,
            binascii.crc32(data) & 0xFFFFFFFF,
            self.SUFFIX,
        )
        return data

    def unpack(
        self, data: bytes
    ) -> Tuple[int, Optional[Response], Optional[Type[Command]], bytes]:
        header_length = struct.calcsize(self.HEADER_FMT)

        # Header.
        if len(data) < header_length:
            # Wait for more data.
            return 0, None, None, data
        header = data[:header_length]
        prefix, sequence_number, cmd, payload_length = struct.unpack(
            self.HEADER_FMT, header
        )
        if prefix != self.PREFIX:
            raise ResponseError(f"incorrect prefix: 0x{prefix:08x}")

        # Payload.
        if len(data) < header_length + payload_length:
            # Wait for more data.
            return 0, None, None, data
        payload = data[header_length : header_length + payload_length]

        if cmd not in self.RESPONSES:
            logger.warning("unknown response type %i, ignoring", cmd)
            return 0, None, None, data[header_length + payload_length :]

        response_factory = self.RESPONSES[cmd]
        command_class = self.COMMAND_CLASSES.get(cmd)

        try:
            # Get the return code.
            return_code_length = struct.calcsize(self.RETURN_CODE_FMT)
            (return_code,) = struct.unpack(
                self.RETURN_CODE_FMT, payload[:return_code_length]
            )

            # Check the suffix, ignore the hash.
            end_length = struct.calcsize(self.END_FMT)
            _, suffix = struct.unpack(self.END_FMT, payload[-end_length:])
            if suffix != self.SUFFIX:
                raise ResponseError(f"incorrect suffix: 0x{suffix:08x}")

            # Parse payload.
            payload_content = payload[return_code_length:-end_length]
            if payload_content.startswith(self._cfg.version):
                payload_content = payload_content[len(self._version_header) :]
            parsed_payload: Optional[Payload] = None
            error: Optional[ResponseError] = None
            if return_code:
                # Use the payload content as the error description.
                error = ResponseError(payload_content.decode())
            elif payload_content:
                try:
                    decrypted = self._cipher.decrypt(payload_content)
                    try:
                        parsed_payload = json.loads(decrypted)
                    except Exception as e:
                        error = DecodeResponseError(
                            f"could not load {decrypted!r} as json", e
                        )
                except Exception as e:
                    error = DecodeResponseError(
                        f"could not decrypt {payload_content!r}", e
                    )

            return (
                sequence_number,
                response_factory(parsed_payload, error),
                command_class,
                data[header_length + payload_length :],
            )
        except Exception as e:
            return (
                sequence_number,
                response_factory(None, ResponseError("error handling response", e)),
                command_class,
                data[header_length + payload_length :],
            )



from typing import Awaitable, Callable

from local_tuya.domoticz.units.base import Unit, UnitCommand, UnitValues


def set_point_unit(
    id_: int,
    name: str,
    image: int,
    command_func: Callable[[float], Awaitable],
) -> Unit[float]:
    def _to_unit_values(value: float) -> UnitValues:
        return UnitValues(n_value=1, s_value=str(round(value, 1)))

    def _command_to_value(command: UnitCommand) -> float:
        return command.level

    return Unit(
        id_=id_,
        type_="Set Point",
        name=name,
        image=image,
        to_unit_values=_to_unit_values,
        command_to_value=_command_to_value,
        command_func=command_func,
    )



from argparse import ArgumentParser
from pathlib import Path
from typing import Optional, Type, TypeVar

from local_tuya.device import State
from local_tuya.domoticz.plugin.metadata import PluginMetadata
from local_tuya.domoticz.plugin.plugin import OnStart
from local_tuya.domoticz.units import UnitId

T = TypeVar("T", bound=State)


def _get_domoticz_path() -> Path:
    parser = ArgumentParser(prog="Domoticz plugin installer")
    parser.add_argument("-p", "--domoticz-path", dest="path", action="store")
    domoticz_path: Optional[str] = parser.parse_args().path
    if domoticz_path:
        return Path(domoticz_path)
    return Path("~/domoticz").expanduser()


def install_plugin(
    metadata: PluginMetadata,
    on_start: OnStart,
    import_path: str,
    unit_ids: Optional[Type[UnitId]] = None,
) -> None:
    target = _get_domoticz_path() / "plugins" / metadata.package / "plugin.py"
    if not target.parent.exists():
        target.parent.mkdir(parents=True)
    template = (Path(__file__).parent / "template.txt").read_text()
    target.write_text(
        template.format(
            definition=metadata.definition(unit_ids),
            package=metadata.package,
            import_path=import_path,
            on_start_name=on_start.__name__,
            unit_ids_name=unit_ids.__name__ if unit_ids else "__NOOP_UNIT_ID__",
        ),
    )



import asyncio
from abc import ABC, abstractmethod
from contextlib import AbstractContextManager


class Backoff(AbstractContextManager, ABC):
    """Can be used to wait according to a strategy until exited."""

    @abstractmethod
    async def __call__(self) -> None:
        """Wait according to the backoff strategy."""


class SequenceBackoff(Backoff):
    """Backoff according to a wait sequence.
    When it reaches the end of the sequence, it will keep using the last value.

    >>> with Backoff(1, 5, 10) as backoff:
    >>>     ...
    >>>     await backoff()
    """

    def __init__(self, *sequence: float):
        self.__seq = tuple(sequence)
        self.__index = 0

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.__index = 0

    async def __call__(self) -> None:
        await asyncio.sleep(self.__seq[self.__index])
        if self.__index < len(self.__seq) - 1:
            self.__index += 1



import asyncio
import logging
from typing import Optional, cast

from concurrent_tasks import BackgroundTask

from local_tuya.backoff import Backoff
from local_tuya.events import EventNotifier
from local_tuya.protocol.events import (
    ConnectionEstablished,
    ConnectionLost,
    DataReceived,
    DataSent,
)

logger = logging.getLogger(__name__)


class Transport(asyncio.Protocol):
    def __init__(
        self,
        address: str,
        port: int,
        backoff: Backoff,
        event_notifier: EventNotifier,
    ):
        event_notifier.register(DataSent, self._write)
        self._notifier = event_notifier
        self._address = address
        self._port = port
        self._transport: Optional[asyncio.transports.WriteTransport] = None
        self._closing = False
        self._closed = asyncio.Event()
        self._closed.set()
        self._close_exc: Optional[Exception] = None
        self._connected = asyncio.Event()
        self._backoff = backoff
        self._connect_task = BackgroundTask(self._connect)
        self._receive_task = BackgroundTask(self._receive)
        # We need a queue as `data_received` is not async.
        self._received_bytes: asyncio.Queue[bytes] = asyncio.Queue()

    async def __aenter__(self):
        self._closed.clear()
        await self._connect_task.create()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        self._closing = True
        self._connect_task.cancel()
        self._receive_task.cancel()
        if self._transport:
            self._transport.close()
            # Wait until `connection_lost` was called.
            try:
                await asyncio.wait_for(self._closed.wait(), timeout=10)
                await self._notifier.emit(ConnectionLost(None))
            except asyncio.TimeoutError:
                logger.error("timeout waiting for transport to close")

    async def _connect(self) -> None:
        if self._close_exc:
            exc = self._close_exc
            self._close_exc = None
            await self._notifier.emit(ConnectionLost(exc))
        self._transport = await self._get_transport()
        self._connected.set()
        logger.info("connected to device %s:%i", self._address, self._port)
        await self._notifier.emit(ConnectionEstablished())
        self._receive_task.create()

    async def _get_transport(self) -> asyncio.transports.WriteTransport:
        with self._backoff as backoff:
            while True:
                try:
                    transport, _ = await asyncio.get_running_loop().create_connection(
                        lambda: self,
                        self._address,
                        self._port,
                    )
                    return cast(asyncio.transports.WriteTransport, transport)
                except Exception:
                    logger.warning("could not connect, retrying...", exc_info=True)
                    await backoff()

    def connection_lost(self, exc: Optional[Exception]) -> None:
        self._close_exc = exc
        if not exc:
            logger.info("disconnected from device %s:%i", self._address, self._port)
            # Notify the transport was properly closed.
            self._closed.set()
        else:
            logger.warning(f"connection lost, {exc}, reconnecting...")
            self._connect_task.cancel()
            self._receive_task.cancel()
            # Attempt to reconnect.
            self._connected.clear()
            self._transport = None
            self._received_bytes = asyncio.Queue()
            self._connect_task.create()

    def data_received(self, data: bytes) -> None:
        self._received_bytes.put_nowait(data)

    async def _receive(self) -> None:
        while True:
            data = await self._received_bytes.get()
            await self._notifier.emit(DataReceived(data))

    async def _write(self, data: DataSent) -> None:
        if self._closed.is_set():
            raise RuntimeError("transport is closed")
        if self._closing:
            raise RuntimeError("transport is closing")
        await self._connected.wait()
        # Writing is asynchronous, errors will be raised through the `connection_lost` method.
        cast(asyncio.transports.WriteTransport, self._transport).write(data)



from dataclasses import dataclass
from typing import Dict, Optional, Union

from typing_extensions import Protocol

from local_tuya.errors import DecodeResponseError, ResponseError

Value = Union[bool, int, str]
Values = Dict[str, Value]
Payload = Dict[str, Union[Value, Values]]


class Command(Protocol):
    @property
    def payload(self) -> Optional[Payload]:
        return None


class Response(Protocol):
    @property
    def error(self) -> Optional[ResponseError]:
        return None


class EmptyCommand:
    @property
    def payload(self) -> Optional[Payload]:
        return None


@dataclass
class EmptyResponse:
    _error: Optional[ResponseError] = None

    @property
    def error(self) -> Optional[ResponseError]:
        return self._error


class HeartbeatCommand(EmptyCommand):
    ...


class HeartbeatResponse(EmptyResponse):
    ...


class StatusResponse:
    def __init__(
        self,
        payload: Optional[Payload] = None,
        error: Optional[ResponseError] = None,
    ):
        self.error = error
        self.values: Values = {}
        if payload and "dps" in payload and isinstance(payload["dps"], dict):
            self.values = payload["dps"]
        if not self.values and not self.error:
            self.error = DecodeResponseError("no dps in response")


class StateCommand(EmptyCommand):
    ...


class StateResponse(StatusResponse):
    ...


@dataclass
class UpdateCommand:
    values: Values

    @property
    def payload(self) -> Payload:
        return {"dps": self.values}


class UpdateResponse(EmptyResponse):
    ...



"""
Minimal Domoticz types for static typing and mocking in tests.
"""
from typing import Dict, Optional


class DomoticzUnit:
    def __init__(
        self,
        Name: str,
        DeviceID: str,
        Unit: int,
        TypeName: str,
        Image: Optional[int] = 0,
        Options: Optional[Dict[str, str]] = None,
    ):
        self.nValue: int = 0
        self.sValue: str = ""
        self.Color: str = ""
        self.Image: Optional[int] = 0
        self.Options: Optional[Dict[str, str]] = None

    def Create(self) -> None:
        ...

    def Update(self, Log: bool) -> None:
        ...

    def Delete(self) -> None:
        ...


class DomoticzEx:
    @classmethod
    def Debug(cls, s: str) -> None:
        ...

    @classmethod
    def Log(cls, s: str) -> None:
        ...

    @classmethod
    def Status(cls, s: str) -> None:
        ...

    @classmethod
    def Error(cls, s: str) -> None:
        ...

    @classmethod
    def Heartbeat(cls, i: int) -> None:
        ...

    @classmethod
    def Debugging(cls, i: int) -> None:
        ...

    Unit = DomoticzUnit


class DomoticzDevice:
    Units: Dict[int, DomoticzUnit] = {}


Parameters: Dict[str, str] = {}
Devices: Dict[str, DomoticzDevice] = {}



import logging
from contextlib import AsyncExitStack
from typing import Any, Awaitable, Callable, Generic, Optional, TypeVar

from local_tuya.device.buffer import UpdateBuffer
from local_tuya.device.config import DeviceConfig
from local_tuya.device.constraints import Constraints
from local_tuya.device.enums import State
from local_tuya.device.state import StateHandler
from local_tuya.events import maybe_async
from local_tuya.protocol import Protocol, Values

logger = logging.getLogger(__name__)
T = TypeVar("T", bound=State)


class Device(AsyncExitStack, Generic[T]):
    def __init__(
        self,
        config: DeviceConfig,
        load_state: Callable[[Values], T],
        state_updated_callback: Optional[Callable[[T], Any]] = None,
        constraints: Optional[Constraints] = None,
    ):
        super().__init__()
        self._load_state = load_state
        self._state_updated_callback_func: Optional[Callable[[T], Awaitable]] = (
            maybe_async(state_updated_callback) if state_updated_callback else None
        )
        self._state_handler = StateHandler(self._state_updated_callback)
        self._protocol = Protocol(
            config.protocol,
            self._state_handler.updated,
        )
        self._buffer = UpdateBuffer(
            delay=config.debounce_updates,
            confirm_timeout=config.confirm_timeout,
            protocol=self._protocol,
            state_handler=self._state_handler,
            constraints=constraints,
        )

    async def __aenter__(self):
        await self.enter_async_context(self._protocol)
        self.enter_context(self._buffer)
        return self

    def set_state_updated_callback(self, callback: Callable[[T], Any]) -> None:
        self._state_updated_callback_func = maybe_async(callback)

    async def _state(self) -> T:
        return self._load_state(await self._state_handler.state())

    async def _update(self, values: Values) -> None:
        await self._buffer.update(values)

    async def _state_updated_callback(self, values: Values) -> None:
        state = self._load_state(values)
        logger.debug("received new device state: %s", state)
        if self._state_updated_callback_func:
            await self._state_updated_callback_func(state)



import asyncio
from contextlib import AsyncExitStack
from typing import Awaitable, Callable, Optional

from local_tuya.events import EventNotifier
from local_tuya.protocol.config import ProtocolConfig
from local_tuya.protocol.events import CommandSent, StateUpdated
from local_tuya.protocol.heartbeat import Heartbeat
from local_tuya.protocol.message import (
    MessageHandler,
    UpdateCommand,
    Values,
    get_handler,
)
from local_tuya.protocol.receiver import Receiver
from local_tuya.protocol.sender import Sender
from local_tuya.protocol.state import State
from local_tuya.protocol.transport import Transport


def _state_updated_callback(
    func: Callable[[Values], Awaitable]
) -> Callable[[StateUpdated], Awaitable]:
    async def _wrapper(event: StateUpdated) -> None:
        await func(event.values)

    return _wrapper


class Protocol(asyncio.Protocol, AsyncExitStack):
    def __init__(
        self,
        config: ProtocolConfig,
        state_updated_callback: Optional[Callable[[Values], Awaitable]] = None,
    ):
        super().__init__()
        self._config = config
        self._event_notifier = EventNotifier()
        if state_updated_callback:
            self._event_notifier.register(
                StateUpdated,
                _state_updated_callback(state_updated_callback),
            )
        message_handler: MessageHandler = get_handler(config)
        self._transport = Transport(
            address=config.address,
            port=config.port,
            backoff=config.connection_backoff,
            event_notifier=self._event_notifier,
        )
        self._sender = Sender(
            message_handler=message_handler,
            event_notifier=self._event_notifier,
            timeout=config.timeout,
        )
        self._receiver = Receiver(
            message_handler=message_handler,
            event_notifier=self._event_notifier,
        )
        self._heartbeat = Heartbeat(
            interval=config.heartbeat_interval,
            event_notifier=self._event_notifier,
        )
        self._state_keeper = State(
            refresh_interval=config.state_refresh_interval,
            event_notifier=self._event_notifier,
        )

    async def __aenter__(self):
        await self.enter_async_context(self._transport)
        await self.enter_async_context(self._sender)
        self.enter_context(self._heartbeat)
        self.enter_context(self._state_keeper)
        return self

    async def update(self, values: Values) -> None:
        """Update the device."""
        await self._event_notifier.emit(CommandSent(UpdateCommand(values)))



from dataclasses import dataclass
from typing import Optional, Type

from local_tuya.events import Event
from local_tuya.protocol.message import Command, Response, Values


class ConnectionEstablished(Event):
    ...


@dataclass
class ConnectionLost(Event):
    error: Optional[Exception]


class DataSent(bytes, Event):
    ...


class DataReceived(bytes, Event):
    ...


@dataclass
class CommandSent(Event):
    command: Command


@dataclass
class ResponseReceived(Event):
    sequence_number: int
    response: Response
    command_class: Optional[Type[Command]]


@dataclass
class StateUpdated(Event):
    values: Values



from typing import Optional


class LocalTuyaError(Exception):
    """Generic local Tuya error."""


class CommandTimeoutError(LocalTuyaError):
    """The device failed to send a response within the time allowed."""


class ResponseError(LocalTuyaError):
    """Generic error in responses"""

    def __init__(self, message: str = "", cause: Optional[Exception] = None):
        super().__init__(message)
        if cause:
            self.__cause__ = cause


class DecodeResponseError(ResponseError):
    """Error parsing Tuya response."""



from local_tuya.domoticz.plugin import Option, Parameter, PluginMetadata, install_plugin
from local_tuya.domoticz.units import (
    SelectorSwitchStyle,
    Unit,
    UnitId,
    UnitManager,
    compose,
    debounce,
    moving_average,
    selector_switch_unit,
    set_point_unit,
    switch_unit,
    temperature_unit,
)
